#	include	"Simulator.h"

StandardDRAMDevice	::	StandardDRAMDevice ()
	{
	Pages = PageSize = AccessLatency = TransferRate = 0;
	UpStreamComponent = NULL;
	PageArray = NULL;
	Transacting = 0;
	}

StandardDRAMDevice	::	~StandardDRAMDevice ()
	{
	Clear();
	}

bool
StandardDRAMDevice	::	PrintConfiguration (
	int		Format,
	FILE	*fp)
	{
	PrintValue(Format, fp, "AccessLatency", AccessLatency);
	PrintValue(Format, fp, "TransferRate", TransferRate);
	return ( true );
	}

bool
StandardDRAMDevice	::	PrintStatistics (
	int		Format,
	FILE	*fp)
	{
	PrintValue(Format, fp, "Accesses", Accesses);
	PrintValue(Format, fp, "BytesRead", BytesRead);
	PrintValue(Format, fp, "BytesWritten", BytesWritten);
	PrintValue(Format, fp, "FrontSideBusBusyCycles", BusBusyCycles);
	PrintValue(Format, fp, "FrontSideBusIdleCycles", BusIdleCycles);
	PrintPercent(Format, fp, "FrontSideBusUtilization", BusBusyCycles, BusIdleCycles);
	PrintValue(Format, fp, "DRAMBusyCycles", DRAMBusyCycles);
	PrintValue(Format, fp, "DRAMIdleCycles", DRAMIdleCycles);
	PrintPercent(Format, fp, "DRAMUtilization", DRAMBusyCycles, DRAMIdleCycles);
	return ( true );
	}

bool
StandardDRAMDevice	::	Clear()
	{
	uint32	Index;

	if(PageArray)
		{
		for(Index=0;Index<Pages;Index++)
			{
			if(PageArray[Index])
				delete PageArray[Index];
			PageArray[Index] = NULL;
			}
		delete PageArray;
		PageArray = NULL;
		}
	Pages = PageSize = AccessLatency = TransferRate = 0;
	if(UpStreamComponent)
		{
		//UpStreamComponent->Disconnect(this);
		UpStreamComponent = NULL;
		}
	BytesRead = 0;
	BytesWritten = 0;
	Accesses = 0;
	BusBusyCycles = BusIdleCycles = 0;
	DRAMBusyCycles = DRAMIdleCycles = 0;
	return ( true );	
	}

bool
StandardDRAMDevice	::	Initialize(
	uint32	aPages,
	uint32	aPageSize,
	uint32	aAccessLatency,
	uint32	aTransferRate)
	{
	uint32	Index;
	uint32	Index2, Index3;
	
	Clear();

	AccessLatency = aAccessLatency;
	TransferRate = aTransferRate;

	Index = aPageSize>>1;
	Index2 = 1;	
	Index3 = 1;
	while(Index)
		{
		++Index2;
		Index3 = (Index3 << 1)|1;
		Index = Index >> 1;
		}
	PageShift = Index2-1;
	PageMask = Index3 >> 1;
	PageSize = (1<<(PageShift));
	Pages = aPages;

	PageArray = new uint8 *[Pages];
	Index = 0;
	for(Index=0;Index<Pages;Index++)
		PageArray[Index] = NULL;

	//printf("PageMask = %8.8x PageShift = %d PageSize = %d\n",
	//	PageMask, PageShift, PageSize);

	return ( true );
	}

bool
StandardDRAMDevice	::	Connect(
	Component	*c)
	{
	UpStreamComponent = c;
	return ( true );
	}

uint8	*
StandardDRAMDevice	::	GetPage (
	SimAddress	sa)
	{
	sa = sa & 0x7fffffff;
	uint32	Index = (sa >> PageShift);
	//if(Index >= Pages)
	//	{
	//	theSimulator->PrintError("Page: %d > Pages=%d\n", Index, Pages);
	//	return(NULL);
	//	}
	if(!PageArray[Index])
		{
		//printf("New Page for address sa=%8.8x index=%8.8x\n", sa, Index);
		PageArray[Index] = new uint8[PageSize];
		memset(PageArray[Index], 0x00, PageSize);
		}
	return(PageArray[Index]);
	}

uint8
StandardDRAMDevice	::	ReadMemory8(
	SimAddress	sa)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(0xff);

	return(Page[sa&PageMask]);
	}

uint16
StandardDRAMDevice	::	ReadMemory16(
	SimAddress	sa)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(0xffff);

	Page = &Page[sa&PageMask];
	uint16	*Temp = (uint16*)Page;
	return(*Temp);
	}

uint32
StandardDRAMDevice	::	ReadMemory32(
	SimAddress	sa)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(0xffffffff);

	Page = &Page[sa&PageMask];
	uint32	*Temp = (uint32*)Page;
	return(*Temp);
	}

bool
StandardDRAMDevice	::	WriteMemory8(
	SimAddress	sa,
	uint8		Data)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(-1);

	Page[sa&PageMask] = Data;
	return ( true );
	} 

bool
StandardDRAMDevice	::	WriteMemory16(
	SimAddress	sa,
	uint16		Data)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(-1);

	Page = &Page[sa&PageMask];
	uint16	*Temp = (uint16*)Page;
	(*Temp) = Data;
	return ( true );
	}

bool
StandardDRAMDevice	::	WriteMemory32(
	SimAddress	sa,
	uint32		Data)
	{
	uint8	*Page = GetPage(sa);
	if(!Page)	return(-1);

	Page = &Page[sa&PageMask];
	uint32	*Temp = (uint32*)Page;
	(*Temp) = Data;
	return ( true );
	}

int
StandardDRAMDevice	::	SimulateOneClockCycle()
	{
	if(FrontSideBus.Busy())
		++BusBusyCycles;
	else
		++BusIdleCycles;

	if(DRAM.Busy())
		++DRAMBusyCycles;
	else
		++DRAMIdleCycles;

	return(1);
	}

bool
StandardDRAMDevice	::	ScheduleOperation(
	bool		IsWrite,
	SimAddress	sa,
	int			Bytes,
	int			StartAfter,
	int			*UntilStart,
	int			*UntilFinish)
	{
	int			Temp1, Temp2;
	Transacting = StartAfter;
	(*UntilStart) = Transacting;
	// Schedule bus stuff
	if(IsWrite)
		{
		FrontSideBus.RequestNewSchedule(Transacting,
			2*TransferRate + TransferRate*(Bytes>>4),
			0, &Temp1, &Temp2);
		Transacting = Temp2;
		}	
	else
		{
		FrontSideBus.RequestNewSchedule(Transacting,
			2*TransferRate,
			0, &Temp1, &Temp2);
		Transacting = Temp2;
		}

	DRAM.RequestNewSchedule	(Transacting, AccessLatency, 0, &Temp1, &Temp2);

	// if read, post back data.
	if(!IsWrite)
		{
		Transacting = Temp2;
		FrontSideBus.RequestNewSchedule(
			Transacting, TransferRate*(Bytes>>4),
			0, &Temp1, &Temp2);
		Transacting = Temp2;
		}

	(*UntilFinish) = Transacting;

	++Accesses;
	if(IsWrite)
		BytesWritten += Bytes;
	else
		BytesRead += Bytes;
	return(true);
	}

